from typing import Type, Union

from amapy_contents import DockerContent
from amapy_pluggy.plugin import hook_impl
from amapy_pluggy.plugin.object_content import ObjectContent
from amapy_pluggy.storage import StorageData, StorageURL, Transporter
from amapy_pluggy.storage.asset_storage import AssetStorage
from amapy_plugin_gcr.gcr_blob import GcrBlob
from amapy_plugin_gcr.gcr_mixin import GcrMixin
from amapy_plugin_gcr.gcr_url import GcrURL
from amapy_plugin_gcr.transporter.gcr_transporter import GcrTransporter


class GcrStorage(AssetStorage, GcrMixin):
    prefixes = ["gcr.io/", "us.gcr.io/"]
    name = "gcr"

    def allows_proxy(self):
        return True

    def allows_object_add(self):
        return True

    def get_storage_url(self, url_string: str, ignore: str = None) -> StorageURL:
        return GcrURL(url=url_string, ignore=ignore)

    def get_transporter(self) -> Type[Transporter]:
        return GcrTransporter.shared()

    def get_content_class(self) -> Type[ObjectContent]:
        return DockerContent

    def get_object_path(self, asset_root: str, blob: GcrBlob, parent_url: StorageURL) -> str:
        return "-".join([*(blob.name.split("/")), *blob.get_hash()])

    def get_blob(self, url: str) -> GcrBlob:
        gcr_url: GcrURL = GcrURL(url=url)
        data = self.get_gcr_blob(url=gcr_url)
        return GcrBlob(data=data, url_object=gcr_url)

    def blob_exists(self, url_string: str) -> bool:
        raise NotImplementedError

    def list_blobs(self, url: Union[StorageURL, str], ignore: str = None) -> [StorageData]:
        if type(url) is str:
            url = GcrURL(url=url, ignore=ignore)
        data = self.get_gcr_blob(url=url)
        return [GcrBlob(data=data, url_object=url)]

    def delete_blobs(self, url_strings: [str]) -> None:
        # delete not allowed in gcr currently
        raise NotImplementedError


class GcrStoragePlugin:
    @hook_impl
    def asset_storage_get(self) -> Type[AssetStorage]:
        return GcrStorage



from amapy_pluggy.storage import StorageData
from amapy_plugin_gcr.gcr_url import GcrURL


class GcrBlob(StorageData):
    url = None
    host: str = None
    name: str = None
    tag: str = None
    content_type: str = None
    hashes: dict = {}

    def initialize(self, data: dict, url_object: GcrURL) -> None:
        """
        Parameters
        ----------
        data: dict
            The data to initialize the blob with
        url_object: GcrURL
            The URL object that the blob is associated with
        """
        self.name = data.get("name")
        self.tag = data.get("tag")
        self.content_type = data.get("mediaType")
        self.hashes[data.get("hash_type")] = data.get("hash_value")
        self.size = int(data.get("imageSizeBytes"))
        self.host = url_object.host
        self.url = url_object.url_for_blob(host=self.host, name=self.name,
                                           hash_type=data.get("hash_type"),
                                           hash_value=data.get("hash_value"))



from amapy_plugin_gcr.gcr_url import GcrURL


def test_gcr_urls():
    test_cases = [
        {
            "url": "gcr.io/my-test-project/my-test-image@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890",
            "expected": {
                "host": "gcr.io",
                "project": "my-test-project",
                "image": "my-test-image",
                "hash": "sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890",
                "tag": None,
                "is_valid": True
            }
        },
        {
            "url": "us.gcr.io/my-test-project/my-test-image@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890",
            "expected": {
                "host": "us.gcr.io",
                "project": "my-test-project",
                "image": "my-test-image",
                "hash": "sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890",
                "tag": None,
                "is_valid": True
            }
        },
        {
            "url": "gcr.io/my-test-project/my-test-image:latest",
            "expected": {
                "host": "gcr.io",
                "project": "my-test-project",
                "image": "my-test-image",
                "hash": None,
                "tag": "latest",
                "is_valid": True
            }
        },
        {
            "url": "gcr.io/my-test-project/my-test-image",
            "expected": {
                "host": "gcr.io",
                "project": "my-test-project",
                "image": "my-test-image",
                "hash": None,
                "tag": None,
                "is_valid": False
            }
        },
        {
            "url": "gcr.io/my-test-project/my-test-image/",
            "expected": {
                "host": "gcr.io",
                "project": "my-test-project",
                "image": "my-test-image",
                "hash": None,
                "tag": None,
                "is_valid": False
            }
        },
        {
            "url": "gcr.io/my-test-project/",
            "expected": {
                "host": "gcr.io",
                "project": "my-test-project",
                "image": None,
                "hash": None,
                "tag": None,
                "is_valid": False
            }
        },
        {
            "url": "gcr.io/",
            "expected": {
                "host": "gcr.io",
                "project": None,
                "image": None,
                "hash": None,
                "tag": None,
                "is_valid": False
            }
        },
    ]

    for case in test_cases:
        gcr_url = GcrURL(url=case["url"])
        expected = case["expected"]
        assert gcr_url.host == expected["host"]
        assert gcr_url.project == expected["project"]
        assert gcr_url.image == expected["image"]
        assert gcr_url.hash == expected["hash"]
        assert gcr_url.tag == expected["tag"]
        assert gcr_url.is_valid() == expected["is_valid"]
        assert gcr_url.is_remote() == expected["is_valid"]



import json

import requests
from google.auth.transport.requests import Request
from google.oauth2.service_account import Credentials

from amapy_plugin_gcr.gcr_url import GcrURL
from amapy_utils.common import exceptions

URL_FORMAT = "https://{host}/v2/{project}/{image}/tags/list"
AUTH_FORMAT = "Bearer {token}"


class GcrMixin:

    def get_gcr_blob(self, url: GcrURL):
        if not url.is_valid():
            raise exceptions.InvalidStorageURLError(f"{exceptions.InvalidStorageURLError.msg}: {url.url}")

        res: dict = self.fetch_url_data(host=url.host,
                                        project=url.project,
                                        image=url.image)
        # we need to parse and filter because it's a list of objects
        return self.parse_url_response(gcr_url=url, url_data=res)

    def fetch_url_data(self, host: str, project: str, image: str) -> dict:
        url = URL_FORMAT.format(host=host, project=project, image=image)
        headers = {"Authorization": AUTH_FORMAT.format(token=self.get_token())}
        res = requests.get(url=url, headers=headers)
        return res.json()

    def parse_url_response(self, gcr_url: GcrURL, url_data: dict):
        image_data = url_data.get("manifest")
        # this is a dict where keys are the hashes
        if not image_data:
            if self.is_permission_error(url_data):
                raise exceptions.InsufficientCredentialError(msg=f"permission denied to: {gcr_url}",
                                                             data=json.dumps(url_data))
            else:
                exceptions.InvalidObjectSourceError(msg=f"not found: {gcr_url}")
        if gcr_url.tag:
            return self._parse_tag_url_response(url=gcr_url, url_data=url_data)
        else:
            return self._parse_sha_url_response(url=gcr_url, url_data=url_data)

    def _parse_tag_url_response(self, url: GcrURL, url_data: dict):
        self._validate_response(data=url_data, url=url)
        image_data = url_data.get("manifest")
        for hash_name in image_data:
            data: dict = image_data.get(hash_name)
            if data and url.tag in data.get("tag"):
                data['hash_type'], data['hash_value'] = hash_name.split(":")
                data['name'] = url_data.get("name")
                return data

    def _parse_sha_url_response(self, url: GcrURL, url_data: dict):
        self._validate_response(data=url_data, url=url)
        image_data = url_data.get("manifest")
        for hash_name in image_data:
            if hash_name == url.hash:
                found = image_data.get(hash_name)
                found['hash_type'], found['hash_value'] = hash_name.split(":")
                found['name'] = url_data.get("name")
                return found

    def _validate_response(self, data: dict, url: GcrURL):
        # this is a dict where keys are the hashes
        if not data.get("manifest"):
            if self.is_permission_error(data):
                raise exceptions.InsufficientCredentialError(
                    msg=f"{exceptions.InsufficientCredentialError.msg}, permission denied to:{url}")
            else:
                exceptions.InvalidObjectSourceError(msg=f"f{url.url} not found")

    def is_permission_error(self, url_data: dict) -> bool:
        errors = url_data.get("errors", [])
        for error in errors:
            if error.get("code") == "DENIED":
                return True
        return False

    def get_token(self) -> str:
        """
        https://stackoverflow.com/questions/53472429/how-to-get-a-gcp-bearer-token-programmatically-with-python

        Returns
        -------
        str:
            token
        """
        # TODO: return the token from the credentials after auth refactor
        credentials = Credentials.from_service_account_info(self.credentials,
                                                            scopes=['https://www.googleapis.com/auth/cloud-platform'])

        credentials.refresh(Request())
        return credentials.token



import json
from unittest.mock import patch

from amapy_plugin_gcr.gcr_blob import GcrBlob
from amapy_plugin_gcr.gcr_storage import GcrStorage


def test_get_blob(fake_url_data):
    urls = [
        "gcr.io/my-test-project/my-test-image@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890",
        "gcr.io/my-test-project/my-test-image:latest",
    ]

    expected = {
        "name": "my-test-project/my-test-image",
        "size": 2391153464,
        "content_type": "application/vnd.docker.distribution.manifest.v2+json",
        "hashes": {"sha256": "1234567890abcdef1234567890abcdef1234567890abcdef1234567890"},
        "tag": ["latest"],
        "host": "gcr.io",
        "url": "gcr.io/my-test-project/my-test-image@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890",
    }

    with patch("amapy_plugin_gcr.gcr_mixin.GcrMixin.fetch_url_data", return_value=fake_url_data):
        for idx, url in enumerate(urls):
            blob: GcrBlob = GcrStorage.shared().get_blob(url=url)
            for key in expected:
                if type(expected[key]) is dict:
                    assert json.dumps(expected[key]) == json.dumps(getattr(blob, key))
                else:
                    assert expected[key] == getattr(blob, key)


def test_list_blobs(fake_url_data):
    urls = [
        "gcr.io/my-test-project/my-test-image@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890",
        "gcr.io/my-test-project/my-test-image:latest",
    ]
    expected = {
        "name": "my-test-project/my-test-image",
        "size": 2391153464,
        "content_type": "application/vnd.docker.distribution.manifest.v2+json",
        "hashes": {"sha256": "1234567890abcdef1234567890abcdef1234567890abcdef1234567890"},
        "tag": ["latest"],
        "host": "gcr.io",
        "url": "gcr.io/my-test-project/my-test-image@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890",
    }

    with patch("amapy_plugin_gcr.gcr_mixin.GcrMixin.fetch_url_data", return_value=fake_url_data):
        for idx, url in enumerate(urls):
            blob: GcrBlob = GcrStorage.shared().get_blob(url=url)
            for key in expected:
                if type(expected[key]) is dict:
                    assert json.dumps(expected[key]) == json.dumps(getattr(blob, key))
                else:
                    assert expected[key] == getattr(blob, key)



from amapy_pluggy.storage.transporter import Transporter, TransportResource


class GcrTransporter(Transporter):

    def upload(self, resources: [TransportResource]):
        # no upload or download allowed currently
        raise NotImplementedError

    def download(self, resources: [TransportResource]):
        # no upload or download allowed currently
        # todo: use docker to download the image
        raise NotImplementedError

    def copy(self, resources: [TransportResource]):
        # no copy allowed currently
        raise NotImplementedError



from amapy_pluggy.storage import StorageURL

HASH_SEPARATOR = "@"
TAG_SEPARATOR = ":"
PATH_SEPARATOR = "/"


class GcrURL(StorageURL):
    host: str = None
    project: str = None
    image: str = None
    hash: str = None
    tag: str = None

    def parse_url(self, url: str) -> None:
        """Parse a gcr url into various components

        e.g. gcr.io/my-test-project/my-test-image@sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890
        - host: gcr.io
        - project: my-test-project
        - image: my-test-image
        - hash: sha256:1234567890abcdef1234567890abcdef1234567890abcdef1234567890

        Parameters
        ----------
        url: str
            The GCR url to parse
        """
        # sanitize by stripping any leading or trailing slashes
        url = url.strip(PATH_SEPARATOR)
        # note: check for HASH_SEPARATOR first, because the TAG_SEPARATOR is also present in the hash-string
        if HASH_SEPARATOR in url:
            image_part, self.hash = url.split(HASH_SEPARATOR)
        elif TAG_SEPARATOR in url:
            image_part, self.tag = url.split(TAG_SEPARATOR)
        else:
            image_part = url
        parts = image_part.split(PATH_SEPARATOR)
        attrs_map = [
            {"name": "host", "index": 0},
            {"name": "project", "index": 1},
            {"name": "image", "index": 2}
        ]
        for attr in attrs_map:
            idx = attr.get("index")
            val = parts[idx] if len(parts) > idx else None
            setattr(self, attr.get("name"), val)

    def is_valid(self):
        if not (self.hash or self.tag):
            return False
        return bool(self.host and self.project and self.image)

    def is_remote(self):
        return self.is_valid()

    def url_for_blob(self, host: str, name: str, hash_type: str, hash_value: str):
        # todo - ":" is a reserved character in linux so we need to fix this
        return f"{host}/{name}{HASH_SEPARATOR}{hash_type}:{hash_value}"
